from typing import Any, Dict

import msgspec
import pytest
from dsws_client.ds_response import (
    DSDataResponse,
    DSGetDataBundleResponse,
    DSGetDataResponse,
    DSGetTokenResponse,
)


@pytest.fixture(name="token_response_dict")
def token_response_fix() -> Dict[str, Any]:
    """Return an example token response."""
    return {
        "Properties": None,
        "TokenExpiry": "/Date(928149600000+0000)/",
        "TokenValue": "n9IcXCXIq+ApBEe0zLBeTrU11D65F6BB2ED74E3467EA0EBE0C",
    }


def test_token_response(token_response_dict: Dict[str, Any]) -> None:
    """Verify responses can be initialized with field aliases."""
    token_response = msgspec.convert(
        token_response_dict,
        type=DSGetTokenResponse,
    )

    assert token_response.token_value == token_response_dict["TokenValue"]
    assert token_response.token_expiry == token_response_dict["TokenExpiry"]
    assert token_response.properties == token_response_dict["Properties"]


def test_init_get_data_response(timeseries_response: Dict[str, Any]) -> None:
    """Verify responses can be initialized with field aliases."""
    response = msgspec.convert(
        {"DataResponse": timeseries_response},
        type=DSGetDataResponse,
    )

    assert response.data_response == msgspec.convert(
        timeseries_response,
        type=DSDataResponse,
    )


def test_init_get_data_bundle_response(timeseries_response: Dict[str, Any]) -> None:
    """Verify responses can be initialized with field aliases."""
    response = msgspec.convert(
        {"DataResponses": [timeseries_response]},
        type=DSGetDataBundleResponse,
    )

    assert response.data_responses == [
        msgspec.convert(
            timeseries_response,
            type=DSDataResponse,
        )
    ]



from typing import Any, Dict

import pytest


@pytest.fixture()
def timeseries_response() -> Dict[str, Any]:
    """Return an example timeseries response."""
    return {
        "AdditionalResponses": [{"Key": "Frequency", "Value": "D"}],
        "DataTypeNames": [{"Key": "P", "Value": None}],
        "DataTypeValues": [
            {
                "DataType": "P",
                "SymbolValues": [
                    {
                        "Currency": "£",
                        "Symbol": "VOD",
                        "Type": 10,
                        "Value": [92.71, 92.01, 90.79, 89.81, 90.66, 89.68],
                    }
                ],
            }
        ],
        "Dates": [
            "/Date(1681689600000+0000)/",
            "/Date(1681776000000+0000)/",
            "/Date(1681862400000+0000)/",
            "/Date(1681948800000+0000)/",
            "/Date(1682035200000+0000)/",
            "/Date(1682294400000+0000)/",
        ],
        "SymbolNames": [{"Key": "VOD", "Value": "VODAFONE GROUP"}],
        "Tag": None,
    }


@pytest.fixture()
def snapshot_response() -> Dict[str, Any]:
    """Return an example snapshot response."""
    return {
        "AdditionalResponses": None,
        "DataTypeNames": [
            {"Key": "NAME", "Value": "NAME"},
            {"Key": "ISIN", "Value": "ISIN CODE"},
        ],
        "DataTypeValues": [
            {
                "DataType": "NAME",
                "SymbolValues": [
                    {
                        "Currency": "£ ",
                        "Symbol": "VOD",
                        "Type": 6,
                        "Value": "VODAFONE GROUP",
                    },
                    {
                        "Currency": "U$",
                        "Symbol": "U:JPM",
                        "Type": 6,
                        "Value": "JP MORGAN CHASE & CO.",
                    },
                ],
            },
            {
                "DataType": "ISIN",
                "SymbolValues": [
                    {
                        "Currency": "£",
                        "Symbol": "VOD",
                        "Type": 6,
                        "Value": "GB00BH4HKS39",
                    },
                    {
                        "Currency": "U$",
                        "Symbol": "U:JPM",
                        "Type": 6,
                        "Value": "US46625H1005",
                    },
                ],
            },
        ],
        "Dates": ["/Date(1577836800000+0000)/"],
        "SymbolNames": [
            {"Key": "VOD", "Value": "VOD"},
            {"Key": "U:JPM", "Value": "U:JPM"},
        ],
        "Tag": "test",
    }


@pytest.fixture()
def invalid_response(timeseries_response: Dict[str, Any]) -> Dict[str, Any]:
    """Return a response with missing dates."""
    response = timeseries_response.copy()
    response["Dates"] = []
    return response



import os
import urllib.parse
from typing import Any, ClassVar, Dict, Optional

import attrs
import httpx
from dotenv import load_dotenv

from dsws_client.version import __version__


@attrs.define()
class DSWSConfig:
    """Configuration for the DSWS client."""

    path: ClassVar[str] = "/DSWSClient/V1/DSService.svc/rest/"
    host: str = "https://product.datastream.com"
    timeout: int = attrs.field(default=180, converter=int)
    proxy: Optional[str] = None
    ssl_cert: Optional[str] = None
    app_id: str = f"dsws-client-{__version__}"
    data_source: Optional[str] = None
    max_concurrency: int = 1
    debug: bool = attrs.field(default=False, converter=attrs.converters.to_bool)

    def __init__(self, **kwargs: Any) -> None:
        """Load configuration from environment variables."""
        load_dotenv()
        init_dict = {}
        for field in attrs.fields(DSWSConfig):
            # try to get the value from kwargs, then from environment variables
            value = kwargs.get(field.name, os.getenv(field.name.upper()))
            if value is not None:
                init_dict[field.name] = value
        self.__attrs_init__(**init_dict)  # type: ignore[attr-defined]

    @property
    def base_url(self) -> str:
        """Return the base URL."""
        return urllib.parse.urljoin(self.host, self.path)

    @property
    def proxies(self) -> Optional[Dict[str, httpx.HTTPTransport]]:
        """Return the proxies."""
        proxy = httpx.HTTPTransport(proxy=self.proxy)
        return {"https://": proxy, "http://": proxy} if self.proxy else None



import datetime as dt
import random
from typing import List, Optional, Type, Union

import pytest
from dsws_client.ds_request import (
    MAX_DATATYPES_PER_REQUEST,
    MAX_INSTRUMENTS_PER_REQUEST,
    MAX_ITEMS_PER_BUNDLE,
    DSDataRequest,
    DSDataType,
    DSDate,
    DSGetDataBundleRequest,
    DSInstrument,
)
from dsws_client.value_objects import (
    DSDateName,
    DSInstrumentPropertyName,
    DSStringKVPair,
)


@pytest.fixture(name="token")
def token_fix() -> str:
    """Return a fixed token."""
    return "test-token"


def make_snapshot_date(
    date: Union[str, dt.date, DSDateName] = dt.date(2018, 1, 1),
) -> DSDate:
    """Return a snapshot date."""
    return DSDate.construct(
        start=date,
        end=None,
        frequency=None,
        kind=0,
    )


def make_timeseries_date(
    start: Union[str, dt.date, DSDateName] = dt.date(2018, 1, 1),
    end: Union[str, dt.date, DSDateName] = dt.date(2018, 1, 1),
    frequency: str = "D",
) -> DSDate:
    """Return a fixed timeseries date."""
    return DSDate.construct(
        start=start,
        end=end,
        frequency=frequency,
        kind=1,
    )


def random_date() -> DSDate:
    """Return either a snapshot or timeseries date."""
    return make_snapshot_date() if random.random() > 0.5 else make_timeseries_date()


@pytest.fixture(name="snapshot_request")
def data_request_fix(snapshot_date: DSDate) -> DSDataRequest:
    """Return a fixed data request."""
    return DSDataRequest(
        instrument=DSInstrument(value="IBM,MSFT"),
        data_types=[DSDataType(value="P")],
        date=snapshot_date,
    )


def make_data_request(
    date: DSDate,
    instrument: Optional[DSInstrument] = None,
    data_types: Optional[List[DSDataType]] = None,
    tag: Optional[str] = None,
) -> DSDataRequest:
    """Generate a data request."""
    if instrument is None:
        instrument = DSInstrument(value="IBM,MSFT")
    if data_types is None:
        data_types = [DSDataType(value="P")]
    return DSDataRequest(
        instrument=instrument,
        data_types=data_types,
        date=date,
        tag=tag,
    )


@pytest.mark.parametrize("date", [make_snapshot_date(), make_timeseries_date()])
def test_request_instantiation(date: DSDate) -> None:
    """Verify that a request is instantiated properly."""
    data_request = make_data_request(date)

    assert data_request.instrument.identifiers == ["IBM", "MSFT"]


@pytest.mark.parametrize(
    ("instrument", "data_types", "expected"),
    [
        (
            DSInstrument(",".join(["DUMMY"] * (MAX_INSTRUMENTS_PER_REQUEST + 1))),
            [DSDataType("P")],
            ValueError,
        ),
        (
            DSInstrument("DUMMY"),
            [DSDataType("P")] * (MAX_DATATYPES_PER_REQUEST + 1),
            ValueError,
        ),
        (
            DSInstrument(",".join(["DUMMY"] * 40)),
            [DSDataType("P")] * 3,
            ValueError,
        ),
    ],
)
def test_request_validation(
    instrument: DSInstrument,
    data_types: List[DSDataType],
    expected: Type[Exception],
) -> None:
    """Verify that requests are validated properly."""
    with pytest.raises(expected):
        DSDataRequest(
            instrument=instrument,
            data_types=data_types,
            date=random_date(),
        )


def test_bundle_validation(token: str) -> None:
    """Verify that bundles are validated properly."""
    requests = [
        make_data_request(random_date()) for _ in range(MAX_ITEMS_PER_BUNDLE + 1)
    ]

    with pytest.raises(ValueError):  # noqa: PT011
        DSGetDataBundleRequest(data_requests=requests, token_value=token)


# ruff: noqa: ERA001
@pytest.mark.parametrize(
    ("identifiers", "expected", "expected_properties"),
    [
        ("IBM,MSFT", ["IBM", "MSFT"], []),
        (
            ["IBM", "MSFT"],
            ["IBM", "MSFT"],
            [DSStringKVPair(DSInstrumentPropertyName.SYMBOL_SET, True)],
        ),
        # TODO: fix expression splitting
        # (
        #     "PCH#(VOD(P),3M)|E",
        #     ["PCH#(VOD(P),3M)"],
        #     [DSStringKVPair(DSInstrumentPropertyName.EXPRESSION, True)],
        # ),
        (
            "LS&PCOMP|L",
            ["LS&PCOMP"],
            [DSStringKVPair(DSInstrumentPropertyName.INSTRUMENT_LIST, True)],
        ),
    ],
)
def test_construct_instrument(
    identifiers: Union[str, List[str]],
    expected: List[str],
    expected_properties: List[DSStringKVPair],
) -> None:
    """Verify instrument is constructed properly."""
    instrument = DSInstrument.construct(identifiers)

    assert instrument.identifiers == expected
    assert instrument.properties == expected_properties

    # test with return_names
    instrument = DSInstrument.construct(identifiers, return_names=True)

    assert instrument.properties == [
        *expected_properties,
        DSStringKVPair(DSInstrumentPropertyName.RETURN_NAME, True),
    ]



import datetime as dt
import re
from typing import Union


def convert_date(date_str: Union[str, dt.datetime]) -> dt.datetime:
    """Convert a timestamp from Datastream to datetime."""
    if isinstance(date_str, dt.datetime):
        return date_str
    match = re.search(r"[0-9+]+", date_str)
    if match is None:
        msg = f"invalid date string: {date_str}"
        raise ValueError(msg)
    timestamp = match.group(0)
    tzinfo = dt.timezone.utc
    if "+" in timestamp:
        timestamp, offset = timestamp.split("+")
        timedelta = dt.timedelta(hours=int(offset[:2]), minutes=int(offset[2:]))
        tzinfo = dt.timezone(timedelta)
    timestamp_int = int(timestamp)
    return dt.datetime.fromtimestamp(timestamp_int / 1000, tz=tzinfo)



"""The DSWS client."""

import concurrent.futures
import itertools
import logging
import sys
from typing import Any, Dict, Iterator, List, Optional, Type, TypeVar, Union

import httpx
import msgspec

from dsws_client.config import DSWSConfig
from dsws_client.ds_request import (
    DSDataRequest,
    DSDataType,
    DSDate,
    DSGetDataBundleRequest,
    DSGetDataRequest,
    DSGetTokenRequest,
    DSInstrument,
    DSRequest,
    bundle_identifiers,
)
from dsws_client.ds_response import (
    DSGetDataBundleResponse,
    DSGetDataResponse,
    DSGetTokenResponse,
)
from dsws_client.exceptions import (
    InvalidResponseError,
    RequestFailedError,
)
from dsws_client.parse import ParsedResponse, aggregate_responses, responses_to_records
from dsws_client.value_objects import DateType, DSStringKVPair, Token

logger = logging.getLogger(__name__)

ResponseCls = TypeVar("ResponseCls")


class DSWSClient:
    """
    Client for the DSWS web service.

    It can be shared between threads.

    Args:
        username: The DSWS username.
        password: The DSWS password.
        host (default: "https://product.datastream.com"): The DSWS host URL.
        timeout (default: 180): The timeout configuration to use when sending
            requests.
        proxy (optional): A proxy URL where all the traffic should be routed.
        ssl_cert (optional): SSL certificates (a.k.a CA bundle) used to
            verify the identity of requested hosts.
        app_id (default: dsws-client-VERSION): The app ID to use in the
            request.
        data_source (optional): The data source to use in the request.
        max_concurrency (default: 1): The maximum number of concurrent
            requests to make.
        debug (default: False): If True, print the request and response
            data to stdout.
    """

    def __init__(self, username: str, password: str, **kwargs: Any) -> None:
        config = DSWSConfig(**kwargs)
        self._username = username
        self._password = password
        self._session = httpx.Client(
            base_url=config.base_url,
            timeout=config.timeout,
            proxies=config.proxies,  # type: ignore[arg-type]
            verify=config.ssl_cert or True,
            headers={"Content-Type": "application/json"},
        )
        self._max_concurrency = config.max_concurrency
        self._app_id = config.app_id
        self._data_source = config.data_source
        self._debug = config.debug
        self._token: Optional[Token] = None

    @property
    def token(self) -> str:
        """Get a token."""
        if self._token is None or self._token.is_expired:
            self._token = self.fetch_token()
        return self._token.token_value

    def fetch_snapshot_data(  # noqa: PLR0913
        self,
        identifiers: List[str],
        fields: List[str],
        start: Optional[DateType] = None,
        tag: Optional[str] = None,
        *,
        concurrent_requests: bool = False,
    ) -> ParsedResponse:
        """Fetch snapshot data."""
        data_responses = self.fetch_snapshot_data_iter(
            identifiers=identifiers,
            fields=fields,
            start=start,
            tag=tag,
            concurrent_requests=concurrent_requests,
        )
        return aggregate_responses(data_responses)

    def fetch_snapshot_data_iter(  # noqa: PLR0913
        self,
        identifiers: List[str],
        fields: List[str],
        start: Optional[DateType] = None,
        tag: Optional[str] = None,
        *,
        concurrent_requests: bool = False,
    ) -> Iterator[ParsedResponse]:
        """Fetch snapshot data, returning an iterator over responses."""
        request_bundles = self.construct_request_bundles(
            identifiers=identifiers,
            fields=fields,
            start=start,
            end=None,
            frequency=None,
            kind=0,
            tag=tag,
            return_symbol_names=True,
            return_field_names=True,
        )
        responses = self.fetch_all(request_bundles, threaded=concurrent_requests)
        data_responses = itertools.chain.from_iterable(
            response.data_responses for response in responses
        )
        return responses_to_records(data_responses)

    def fetch_timeseries_data(  # noqa: PLR0913
        self,
        identifiers: List[str],
        fields: List[str],
        start: Optional[DateType] = None,
        end: Optional[DateType] = None,
        frequency: str = "D",
        tag: Optional[str] = None,
        *,
        concurrent_requests: bool = False,
    ) -> ParsedResponse:
        """Fetch timeseries data."""
        data_responses = self.fetch_timeseries_data_iter(
            identifiers=identifiers,
            fields=fields,
            start=start,
            end=end,
            frequency=frequency,
            tag=tag,
            concurrent_requests=concurrent_requests,
        )
        return aggregate_responses(data_responses)

    def fetch_timeseries_data_iter(  # noqa: PLR0913
        self,
        identifiers: List[str],
        fields: List[str],
        start: Optional[DateType] = None,
        end: Optional[DateType] = None,
        frequency: str = "D",
        tag: Optional[str] = None,
        *,
        concurrent_requests: bool = False,
    ) -> Iterator[ParsedResponse]:
        """Fetch timeseries data, returning an iterator over responses."""
        request_bundles = self.construct_request_bundles(
            identifiers=identifiers,
            fields=fields,
            start=start,
            end=end,
            frequency=frequency,
            kind=1,
            tag=tag,
            return_symbol_names=True,
            return_field_names=True,
        )
        responses = self.fetch_all(request_bundles, threaded=concurrent_requests)
        data_responses = itertools.chain.from_iterable(
            response.data_responses for response in responses
        )
        return responses_to_records(data_responses)

    def fetch_one(
        self,
        data_request: DSDataRequest,
        **kwargs: object,
    ) -> DSGetDataResponse:
        """
        Post a data request.

        Args:
            data_request: A data request.
            **kwargs: Additional properties to set on the request.

        Returns:
            A data response.
        """
        logger.debug("fetching one")
        return self._execute_request(
            DSGetDataRequest(
                token_value=self.token,
                data_request=data_request,
                properties=[
                    DSStringKVPair(key, value) for key, value in kwargs.items()
                ],
            ),
            DSGetDataResponse,
        )

    def fetch_bundle(
        self,
        data_requests: List[DSDataRequest],
        **kwargs: object,
    ) -> DSGetDataBundleResponse:
        """
        Post multiple data requests.

        Args:
            data_requests: A list of data requests.
            **kwargs: Additional properties to set on the request.

        Returns:
            A data bundle response.
        """
        logger.debug("fetching bundle")
        return self._execute_request(
            DSGetDataBundleRequest(
                token_value=self.token,
                data_requests=data_requests,
                properties=[
                    DSStringKVPair(key, value) for key, value in kwargs.items()
                ],
            ),
            DSGetDataBundleResponse,
        )

    def fetch_all(
        self,
        request_bundles: List[List[DSDataRequest]],
        *,
        threaded: bool = False,
    ) -> Iterator[DSGetDataBundleResponse]:
        """Fetch as many bundles as needed to get all items."""
        if threaded:
            yield from self.fetch_all_threaded(request_bundles)
        else:
            for bundle in request_bundles:
                yield self.fetch_bundle(bundle)

    def fetch_all_threaded(
        self,
        request_bundles: List[List[DSDataRequest]],
    ) -> Iterator[DSGetDataBundleResponse]:
        """Fetch as many bundles as needed to get all items (concurrently)."""
        with concurrent.futures.ThreadPoolExecutor(
            max_workers=self._max_concurrency
        ) as executor:
            logger.debug("fetching bundles in parallel")
            futures = [
                executor.submit(self.fetch_bundle, bundle) for bundle in request_bundles
            ]
            for future in concurrent.futures.as_completed(futures):
                yield future.result()

    def fetch_token(self, **kwargs: object) -> Token:
        """
        Fetch a new token.

        Args:
            **kwargs: Additional properties to set on the request.

        Returns:
            A token.
        """
        token_response = self._execute_request(
            DSGetTokenRequest(
                self._username,
                self._password,
                properties=[
                    DSStringKVPair(key, value) for key, value in kwargs.items()
                ],
            ),
            DSGetTokenResponse,
        )
        return token_response.to_token()

    def construct_request(  # noqa: PLR0913
        self,
        identifiers: Union[str, List[str]],
        fields: List[str],
        start: Optional[DateType],
        end: Optional[DateType],
        frequency: Optional[str],
        kind: int,
        tag: Optional[str] = None,
        *,
        return_symbol_names: bool = False,
        return_field_names: bool = False,
        instrument_props: Optional[Dict[str, str]] = None,
        field_props: Optional[Dict[str, str]] = None,
    ) -> DSDataRequest:
        """Construct a data request."""
        instrument = DSInstrument.construct(
            identifiers,
            return_names=return_symbol_names,
            properties=instrument_props,
        )
        data_types = [
            DSDataType.construct(
                field,
                return_name=return_field_names,
                properties=field_props,
            )
            for field in fields
        ]
        date = DSDate.construct(start, end, frequency, kind)
        return DSDataRequest(instrument, data_types, date, tag)

    def construct_request_bundles(  # noqa: PLR0913
        self,
        identifiers: List[str],
        fields: List[str],
        start: Optional[DateType],
        end: Optional[DateType],
        frequency: Optional[str],
        kind: int,
        tag: Optional[str] = None,
        *,
        return_symbol_names: bool = False,
        return_field_names: bool = False,
        instrument_props: Optional[Dict[str, str]] = None,
        field_props: Optional[Dict[str, str]] = None,
    ) -> List[List[DSDataRequest]]:
        """Construct a list of data request bundles."""
        instrument = DSInstrument.construct(
            identifiers,
            return_names=return_symbol_names,
            properties=instrument_props,
        )
        data_types = [
            DSDataType.construct(
                field,
                return_name=return_field_names,
                properties=field_props,
            )
            for field in fields
        ]
        date = DSDate.construct(start, end, frequency, kind)
        identifier_bundles = bundle_identifiers(instrument, len(data_types))
        request_bundles = []
        for identifier_bundle in identifier_bundles:
            data_requests = []
            for instrument in identifier_bundle:
                data_requests.append(
                    DSDataRequest(instrument, data_types, date, tag=tag)
                )
            request_bundles.append(data_requests)
        return request_bundles

    def _execute_request(
        self,
        request: DSRequest,
        response_cls: Type[ResponseCls],
    ) -> ResponseCls:
        """Execute a request."""
        logger.debug("executing request")
        self._prep_request(request)
        request_data = msgspec.json.encode(request)
        if self._debug:
            sys.stdout.write(f"sending request: {request_data!s}")
        response = self._session.post(
            request.path,
            content=request_data,
        )
        if not response.is_success:
            msg = f"request failed: {response.text}"
            raise RequestFailedError(msg, response.status_code)
        try:
            response_decoded = msgspec.json.decode(response.content, type=response_cls)
        except (msgspec.ValidationError, ValueError, TypeError) as exc:
            msg = f"invalid response: {response.text}"
            raise InvalidResponseError(msg) from exc
        if self._debug:
            sys.stdout.write(f"received response: {response_decoded!s}")
        return response_decoded

    def _prep_request(self, request: DSRequest) -> None:
        """Prepare a request."""
        if self._app_id is not None:
            request.properties.append(DSStringKVPair("__AppId", self._app_id))
        if self._data_source is not None:
            request.properties.append(DSStringKVPair("Source", self._data_source))



class DSWSError(Exception):
    """Base class for DSWS exceptions."""


class RequestFailedError(DSWSError):
    """Indicates that a request failed."""

    def __init__(self, message: str, status_code: int) -> None:
        """Initialize the exception."""
        super().__init__(message)
        self.status_code = status_code


class InvalidResponseError(DSWSError):
    """Indicates that the response is invalid."""


class InvalidRequestError(DSWSError):
    """Indicates that the request is invalid."""



"""DSWS value objects."""

__all__ = (
    # objects
    "DSStringKVPair",
    "Token",
    # enums
    "DSDateFrequencyName",
    "DSDateKind",
    "DSDateName",
    "DSInstrumentPropertyName",
    "DSSymbolResponseValueType",
    # type aliases
    "DateType",
)

from dsws_client.value_objects.enums import (
    DSDateFrequencyName,
    DSDateKind,
    DSDateName,
    DSInstrumentPropertyName,
    DSSymbolResponseValueType,
)
from dsws_client.value_objects.types import DateType, DSStringKVPair, Token



"""DSWS responses."""

import datetime as dt
import sys
from typing import Any, List, Optional, Union

import msgspec

from dsws_client import converters
from dsws_client.value_objects import (
    DSStringKVPair,
    Token,
    enums,
)

if sys.version_info >= (3, 9):
    from typing import Annotated
else:
    from typing_extensions import Annotated

BOOL_MAPPING = {
    "Y": True,
    "true": True,
    "True": True,
    "N": False,
    "false": False,
    "False": False,
}

DSDateString = Annotated[
    str,
    msgspec.Meta(pattern=r"\/Date\((\d+)([+-]\d{4})?\)\/"),
]


class _DSSymbolResponseValue(msgspec.Struct, rename="pascal", tag_field="Type"):
    """A symbol response value."""

    symbol: str
    currency: Optional[str]
    value: Any

    def parse(self) -> Any:
        """Parse the value."""
        return self.value


class DSError(_DSSymbolResponseValue, tag=enums.DSSymbolResponseValueType.ERROR.value):
    """An error response value."""

    value: str

    def parse(self) -> None:
        """Parse the value."""
        return


class DSEmpty(_DSSymbolResponseValue, tag=enums.DSSymbolResponseValueType.EMPTY.value):
    """An empty response value."""

    value: Any

    def parse(self) -> None:
        """Parse the value."""
        return


class DSBool(_DSSymbolResponseValue, tag=enums.DSSymbolResponseValueType.BOOL.value):
    """A boolean response value."""

    value: str

    def parse(self) -> bool:
        """Parse the value."""
        return BOOL_MAPPING[self.value]


class DSInt(_DSSymbolResponseValue, tag=enums.DSSymbolResponseValueType.INT.value):
    """An integer response value."""

    value: int


class DSDateTime(
    _DSSymbolResponseValue,
    tag=enums.DSSymbolResponseValueType.DATE_TIME.value,
):
    """A date/time response value."""

    value: DSDateString

    def parse(self) -> dt.datetime:
        """Parse the value."""
        return converters.convert_date(self.value)


class DSDouble(
    _DSSymbolResponseValue,
    tag=enums.DSSymbolResponseValueType.DOUBLE.value,
):
    """A double response value."""

    value: float


class DSString(
    _DSSymbolResponseValue,
    tag=enums.DSSymbolResponseValueType.STRING.value,
):
    """A string response value."""

    value: str


class DSBoolArray(
    _DSSymbolResponseValue,
    tag=enums.DSSymbolResponseValueType.BOOL_ARRAY.value,
):
    """A boolean array response value."""

    value: str

    def parse(self) -> List[bool]:
        """Parse the value."""
        return [BOOL_MAPPING[item] for item in self.value]


class DSIntArray(
    _DSSymbolResponseValue,
    tag=enums.DSSymbolResponseValueType.INT_ARRAY.value,
):
    """An integer array response value."""

    value: List[int]


class DSDateTimeArray(
    _DSSymbolResponseValue,
    tag=enums.DSSymbolResponseValueType.DATE_TIME_ARRAY.value,
):
    """A date/time array response value."""

    value: List[DSDateString]

    def parse(self) -> List[dt.datetime]:
        """Parse the value."""
        return [converters.convert_date(date_str) for date_str in self.value]


class DSDoubleArray(
    _DSSymbolResponseValue,
    tag=enums.DSSymbolResponseValueType.DOUBLE_ARRAY.value,
):
    """A double array response value."""

    value: List[float]


class DSStringArray(
    _DSSymbolResponseValue,
    tag=enums.DSSymbolResponseValueType.STRING_ARRAY.value,
):
    """A string array response value."""

    value: List[str]


class DSObjectArray(
    _DSSymbolResponseValue,
    tag=enums.DSSymbolResponseValueType.OBJECT_ARRAY.value,
):
    """An object array response value."""

    value: List[Any]


DSSymbolResponseValue = Union[
    DSError,
    DSEmpty,
    DSBool,
    DSInt,
    DSDateTime,
    DSDouble,
    DSString,
    DSBoolArray,
    DSIntArray,
    DSDateTimeArray,
    DSDoubleArray,
    DSStringArray,
    DSObjectArray,
]


class DSDataTypeResponseValue(msgspec.Struct, rename="pascal"):
    """A data type response value."""

    data_type: str
    symbol_values: List[DSSymbolResponseValue]


class DSDataResponse(msgspec.Struct, rename="pascal"):
    """Object that contains a data response."""

    data_type_values: List[DSDataTypeResponseValue]
    dates: Optional[List[DSDateString]]
    data_type_names: Optional[List[DSStringKVPair]]
    symbol_names: Optional[List[DSStringKVPair]]
    additional_responses: Optional[List[DSStringKVPair]]
    tag: Optional[str]

    def pydates(self) -> Optional[List[dt.datetime]]:
        """Return response dates as python dates."""
        if self.dates is None:
            return None
        return [converters.convert_date(date_str) for date_str in self.dates]


class DSGetDataResponse(msgspec.Struct, rename="pascal"):
    """Object that containes a get data response."""

    data_response: DSDataResponse
    properties: Optional[List[DSStringKVPair]] = None


class DSGetDataBundleResponse(msgspec.Struct, rename="pascal"):
    """Object that contains a get data bundle response."""

    data_responses: List[DSDataResponse]
    properties: Optional[List[DSStringKVPair]] = None


class DSGetTokenResponse(msgspec.Struct, rename="pascal"):
    """Object that contains a token response."""

    token_value: str
    token_expiry: DSDateString
    properties: Optional[List[DSStringKVPair]]

    def to_token(self) -> Token:
        """Convert to a token."""
        return Token(self.token_value, converters.convert_date(self.token_expiry))



import datetime as dt
import os

import pytest
from dotenv import load_dotenv
from dsws_client import DSWSClient

load_dotenv()


def has_credentials() -> bool:
    """Return True if credentials are available."""
    return (
        os.environ.get("DSWS_USERNAME") is not None
        and os.environ.get("DSWS_PASSWORD") is not None
    )


@pytest.fixture(name="client")
def client_fix() -> DSWSClient:
    """Provide a DSWSClient instance."""
    username = os.getenv("DSWS_USERNAME", "")
    password = os.getenv("DSWS_PASSWORD", "")
    return DSWSClient(username, password)


@pytest.fixture(name="tag")
def tag_fix() -> str:
    """Provide a test tag."""
    return "test"


@pytest.mark.skipif(not has_credentials(), reason="no credentials available")
def test_get_token(client: DSWSClient) -> None:
    """Test get_token method."""
    token_response = client.fetch_token()

    assert token_response.token_value is not None
    assert token_response.token_expiry is not None
    assert not token_response.is_expired


@pytest.mark.skipif(not has_credentials(), reason="no credentials available")
def test_fetch_snapshot_data(client: DSWSClient, tag: str) -> None:
    """Test fetch_snapshot_data method."""
    identifiers = ["VOD", "U:JPM"]
    fields = ["NAME", "ISIN"]
    start = dt.date(2020, 1, 1)
    response = client.fetch_snapshot_data(
        identifiers=identifiers,
        fields=fields,
        start=start,
        tag=tag,
    )

    assert len(response.records) == 2
    assert len(response.errors) == 0
    assert len(response.meta.data_type_names) == 2
    assert len(response.meta.symbol_names) == 2
    assert len(response.meta.additional_responses) == 0
    assert len(response.meta.tags) == 1
    assert all(isinstance(record["date"], dt.datetime) for record in response.records)


@pytest.mark.skipif(not has_credentials(), reason="no credentials available")
def test_fetch_timeseries_data(client: DSWSClient, tag: str) -> None:
    """Test fetch_timeseries_data method."""
    identifiers = ["VOD", "U:JPM"]
    fields = ["P", "MV"]
    start = dt.date(2020, 1, 1)
    end = dt.date(2020, 1, 10)
    response = client.fetch_timeseries_data(
        identifiers=identifiers,
        fields=fields,
        start=start,
        end=end,
        tag=tag,
    )

    assert len(response.records) == 16
    assert len(response.errors) == 0
    assert len(response.meta.data_type_names) == 2
    assert len(response.meta.symbol_names) == 2
    assert len(response.meta.additional_responses) == 1
    assert len(response.meta.tags) == 1



"""DSWS requests."""

import datetime as dt
import sys
from typing import (
    Any,
    ClassVar,
    Dict,
    List,
    Optional,
    Protocol,
    Tuple,
    TypeVar,
    Union,
)

import msgspec

from dsws_client.value_objects import DSStringKVPair, enums

if sys.version_info >= (3, 9):
    from typing import Annotated
else:
    from typing_extensions import Annotated

# request limits
MAX_INSTRUMENTS_PER_REQUEST = 50
MAX_DATATYPES_PER_REQUEST = 50
MAX_ITEMS_PER_REQUEST = 100

# bundle limits
MAX_REQUESTS_PER_BUNDLE = 20
MAX_ITEMS_PER_BUNDLE = 500


class DSRequest(Protocol):
    """A DSWS request."""

    path: ClassVar[str]
    properties: List[DSStringKVPair]


_HINT_MAP = {
    "L": DSStringKVPair(enums.DSInstrumentPropertyName.INSTRUMENT_LIST, True),
    "E": DSStringKVPair(enums.DSInstrumentPropertyName.EXPRESSION, True),
}


class DSGetTokenRequest(msgspec.Struct, rename="pascal"):
    """Object that defines a token request."""

    path: ClassVar[str] = "GetToken"
    user_name: str
    password: str
    properties: List[DSStringKVPair] = msgspec.field(default_factory=list)

    def add_property(self, key: str, value: str) -> None:
        """Add a property to the request."""
        self.properties.append(DSStringKVPair(key, value))


class DSInstrument(msgspec.Struct, rename="pascal"):
    """Object that defines an instrument."""

    value: str
    properties: List[DSStringKVPair] = msgspec.field(default_factory=list)

    @property
    def identifiers(self) -> List[str]:
        """Return the identifiers."""
        # TODO: this breaks for expressions
        return self.value.split(",")

    @classmethod
    def from_list(cls, instrument_list: List[str]) -> "DSInstrument":
        """Return an instrument from a list of instrument names."""
        properties = [DSStringKVPair(enums.DSInstrumentPropertyName.SYMBOL_SET, True)]
        return cls(",".join(instrument_list), properties)

    @classmethod
    def construct(
        cls,
        identifiers: Union[str, List[str]],
        return_names: bool = False,
        properties: Optional[Dict[str, str]] = None,
    ) -> "DSInstrument":
        """Return an instrument."""
        if isinstance(identifiers, list):
            instance = cls.from_list(identifiers)
        else:
            value = identifiers
            hint_properties = []
            if "|" in value:
                value, hint = identifiers.split("|")
                hint_properties.append(_HINT_MAP[hint])
            instance = cls(value, hint_properties)
        if return_names:
            instance.properties.append(
                DSStringKVPair(enums.DSInstrumentPropertyName.RETURN_NAME, True)
            )
        properties = properties or {}
        for key, value in properties.items():
            instance.properties.append(DSStringKVPair(key, value))
        return instance


class DSDataType(msgspec.Struct, rename="pascal"):
    """A data type (field)."""

    value: str
    properties: List[DSStringKVPair] = msgspec.field(default_factory=list)

    @classmethod
    def construct(
        cls,
        field: str,
        return_name: bool = False,
        properties: Optional[Dict[str, str]] = None,
    ) -> "DSDataType":
        """Construct a data type."""
        instance = cls(field)
        if return_name:
            instance.properties.append(DSStringKVPair("ReturnName", True))
        properties = properties or {}
        for key, value in properties.items():
            instance.properties.append(DSStringKVPair(key, value))
        return instance


class DSDate(msgspec.Struct, rename="pascal"):
    """Date information."""

    start: str
    end: str
    frequency: Optional[enums.DSDateFrequencyName]
    kind: enums.DSDateKind

    @classmethod
    def construct(
        cls,
        start: Optional[Union[str, dt.date, enums.DSDateName]],
        end: Optional[Union[str, dt.date, enums.DSDateName]],
        frequency: Optional[str],
        kind: int,
    ) -> "DSDate":
        """Construct a date."""
        return cls(
            cls._convert_date(start),
            cls._convert_date(end),
            None if frequency is None else enums.DSDateFrequencyName(frequency),
            enums.DSDateKind(kind),
        )

    @classmethod
    def _convert_date(
        cls,
        date: Optional[Union[str, dt.date, enums.DSDateName]],
    ) -> str:
        """Convert a date to a string."""
        if date is None:
            return ""
        if isinstance(date, dt.date):
            return date.isoformat()
        if isinstance(date, str):
            date = enums.DSDateName(date)
        return date.value


class DSDataRequest(msgspec.Struct, rename="pascal"):
    """Object that defines a data request."""

    instrument: DSInstrument
    data_types: Annotated[List[DSDataType], msgspec.Meta(min_length=1)]
    date: DSDate
    tag: Optional[str] = None

    def __post_init__(self) -> None:
        """Validate that a request complies with DSWS request limits."""
        if len(self.instrument.identifiers) > MAX_INSTRUMENTS_PER_REQUEST:
            msg = (
                f"Request contains more than {MAX_INSTRUMENTS_PER_REQUEST} instruments."
            )
            raise ValueError(msg)
        if len(self.data_types) > MAX_DATATYPES_PER_REQUEST:
            msg = f"Request contains more than {MAX_DATATYPES_PER_REQUEST} data types."
            raise ValueError(msg)
        if (
            len(self.instrument.identifiers) * len(self.data_types)
            > MAX_ITEMS_PER_REQUEST
        ):
            msg = f"Request contains more than {MAX_ITEMS_PER_REQUEST} items."
            raise ValueError(msg)


class DSGetDataRequest(msgspec.Struct, rename="pascal"):
    """Object that contains a data request."""

    path: ClassVar[str] = "GetData"
    token_value: str
    data_request: DSDataRequest
    properties: List[DSStringKVPair] = msgspec.field(default_factory=list)

    def add_property(self, key: str, value: str) -> None:
        """Add a property to the request."""
        self.properties.append(DSStringKVPair(key, value))


class DSGetDataBundleRequest(msgspec.Struct, rename="pascal"):
    """Object that contains multiple data requests."""

    path: ClassVar[str] = "GetDataBundle"
    token_value: str
    data_requests: Annotated[List[DSDataRequest], msgspec.Meta(min_length=1)]
    properties: List[DSStringKVPair] = msgspec.field(default_factory=list)

    def __post_init__(self) -> None:
        """Validate that a bundle complies with DSWS bundle limits."""
        if len(self.data_requests) > MAX_REQUESTS_PER_BUNDLE:
            msg = f"Bundle contains more than {MAX_REQUESTS_PER_BUNDLE} requests."
            raise ValueError(msg)
        total_items = sum(
            len(request.instrument.identifiers) * len(request.data_types)
            for request in self.data_requests
        )
        if total_items > MAX_ITEMS_PER_BUNDLE:
            msg = f"Bundle contains more than {MAX_ITEMS_PER_BUNDLE} items."
            raise ValueError(msg)

    def add_property(self, key: str, value: str) -> None:
        """Add a property to the request."""
        self.properties.append(DSStringKVPair(key, value))


T = TypeVar("T", bound=msgspec.Struct)


def evolve(instance: T, **changes: object) -> T:
    """Evolve an instance."""

    def get_field(name: str) -> Any:
        """Return the field."""
        return changes.get(name, getattr(instance, name))

    return instance.__class__(
        **{field: get_field(field) for field in instance.__struct_fields__}
    )


def bundle_identifiers(
    instrument: DSInstrument,
    n_fields: int,
) -> List[Tuple[DSInstrument, ...]]:
    """
    Make a list of bundles of identifiers.

    Args:
        instrument: A DSInstrument to bundle identifiers for.
        n_fields: Number of fields in the request.

    Returns:
        List of bundles of identifiers.

    Raises:
        ValueError: If the number of fields exceeds the maximum allowed.
    """
    if n_fields > MAX_DATATYPES_PER_REQUEST:
        raise ValueError("Number of fields exceeds maximum allowed.")
    identifiers = instrument.identifiers
    bundles = []
    n_identifiers_per_batch = MAX_ITEMS_PER_BUNDLE // n_fields
    for batch_idx in range(0, len(identifiers), n_identifiers_per_batch):
        batch_identifiers = identifiers[batch_idx : batch_idx + n_identifiers_per_batch]
        bundles.append(
            split_bundle_identifiers(
                evolve(instrument, value=",".join(batch_identifiers)),
                n_fields,
            )
        )
    return bundles


def split_bundle_identifiers(
    bundle_instrument: DSInstrument,
    n_fields: int,
) -> Tuple[DSInstrument, ...]:
    """
    Split identifiers per bundle identifiers per bundle request.

    Args:
        bundle_instrument: A DSInstrument to split.
        n_fields: Number of fields in the request.

    Returns:
        Tuple of DSInstruments.
    """
    identifiers = bundle_instrument.identifiers
    request_bundles = []
    n_identifiers_per_request = min(
        MAX_INSTRUMENTS_PER_REQUEST,
        MAX_ITEMS_PER_REQUEST // n_fields,
    )
    for request_idx in range(0, len(identifiers), n_identifiers_per_request):
        bundle_request_identifiers = identifiers[
            request_idx : request_idx + n_identifiers_per_request
        ]
        request_bundles.append(
            evolve(bundle_instrument, value=",".join(bundle_request_identifiers))
        )
    return tuple(request_bundles)



import enum


class DSDateName(str, enum.Enum):
    """Specifies the date literals."""

    BASE_DATE = "BDATE"
    START_OF_WEEK = "SWDATE"
    START_OF_MONTH = "SMDATE"
    START_OF_QUARTER = "SQDATE"
    START_OF_YEAR = "SYDATE"
    LATEST_DATE = "LATESTDATE"
    INTRADAY = "TODAY"
    END_OF_YEAR = "YRE"
    END_OF_MONTH = "MTE"
    END_OF_QUARTER = "QTE"
    END_OF_WEEK = "WKE"
    START_OF_YEAR_2 = "YRS"
    START_OF_MONTH_2 = "MTS"
    START_OF_QUARTER_2 = "QTS"
    START_OF_WEEK_2 = "WKS"
    MIDDLE_OF_YEAR = "YRM"
    MIDDLE_OF_MONTH = "MTM"
    MIDDLE_OF_QUARTER = "QTM"
    MIDDLE_OF_WEEK = "WKM"


class DSDateFrequencyName(str, enum.Enum):
    """Specifies the frequency values."""

    DAILY = "D"
    WEEKLY = "W"
    MONTHLY = "M"
    QUARTERLY = "Q"
    YEARLY = "Y"
    SEVEN_DAYS = "7D"
    SEVEN_DAYS_PADDED = "7DPAD"
    SEVEN_DAYS_MIDDLE_EASTERN = "7DME"
    FIVSEVEN_DAYS_MIDDLE_EASTERN_PADDED = "7DMEPAD"


class DSDateKind(int, enum.Enum):
    """Specifies the date kind."""

    SNAPSHOT = 0
    TIMESERIES = 1


class DSInstrumentPropertyName(str, enum.Enum):
    """Specifies the instrument properties."""

    INSTRUMENT_LIST = "IsList"
    EXPRESSION = "IsExpression"
    SYMBOL_SET = "IsSymbolSet"
    RETURN_NAME = "ReturnName"


class DSSymbolResponseValueType(int, enum.Enum):
    """Specifies the value type of a symbol response."""

    ERROR = 0
    EMPTY = 1
    BOOL = 2
    INT = 3
    DATE_TIME = 4
    DOUBLE = 5
    STRING = 6
    BOOL_ARRAY = 7
    INT_ARRAY = 8
    DATE_TIME_ARRAY = 9
    DOUBLE_ARRAY = 10
    STRING_ARRAY = 11
    OBJECT_ARRAY = 12



from typing import Any, Dict, Optional, Union

import msgspec
import pytest
from dsws_client.ds_response import DSDataResponse, DSDoubleArray
from dsws_client.exceptions import InvalidResponseError
from dsws_client.parse import parse_response, process_string_value, process_symbol_value


def test_invalid_response(invalid_response: Dict[str, Any]) -> None:
    """Verify parsing an invalid response raises an error."""
    response = msgspec.convert(invalid_response, type=DSDataResponse)

    with pytest.raises(InvalidResponseError):
        parse_response(response)


def test_process_response_invalid_value() -> None:
    """Verify processing a value raises an exception if invalid."""
    symbol_value = DSDoubleArray(
        symbol="AAPL",
        currency="E",
        value=[1, 2, 3],
    )

    with pytest.raises(InvalidResponseError):
        process_symbol_value({}, [], "field", [], symbol_value, process_strings=False)


@pytest.mark.parametrize(
    ("value", "expected"),
    [("NA", None), ("N", False), ("Y", True), ("TEST", "TEST")],
)
def test_process_string_value(value: str, expected: Optional[Union[str, bool]]) -> None:
    """Verify strings are cast to correct type."""
    assert process_string_value(value) == expected



import collections
import datetime as dt
import logging
from typing import Any, Dict, Iterable, Iterator, List, Optional, Tuple, Union

import msgspec

from dsws_client.ds_response import (
    DSDataResponse,
    DSError,
    DSString,
    DSSymbolResponseValue,
)
from dsws_client.exceptions import InvalidResponseError

logger = logging.getLogger(__name__)

RecordDict = Dict[Tuple[str, dt.datetime], Dict[str, Any]]


class Error(msgspec.Struct):
    """Error object."""

    field: str
    symbol: str
    message: str


class Meta(msgspec.Struct):
    """Meta object."""

    data_type_names: Dict[str, str] = msgspec.field(default_factory=dict)
    symbol_names: Dict[str, str] = msgspec.field(default_factory=dict)
    additional_responses: Dict[str, str] = msgspec.field(default_factory=dict)
    tags: List[str] = msgspec.field(default_factory=list)
    currencies: Dict[str, Dict[str, Optional[str]]] = msgspec.field(
        default_factory=lambda: collections.defaultdict(dict)
    )

    def merge(self, other: "Meta") -> "Meta":
        """Merge this meta object with another one."""
        return Meta(
            data_type_names={**self.data_type_names, **other.data_type_names},
            symbol_names={**self.symbol_names, **other.symbol_names},
            additional_responses={
                **self.additional_responses,
                **other.additional_responses,
            },
            tags=list({*self.tags, *other.tags}),
            currencies={**self.currencies, **other.currencies},
        )


class ParsedResponse(msgspec.Struct):
    """Parsed response object."""

    records: List[Dict[str, Any]] = msgspec.field(default_factory=list)
    errors: List[Error] = msgspec.field(default_factory=list)
    meta: Meta = msgspec.field(default_factory=Meta)


def responses_to_records(
    responses: Iterable[DSDataResponse],
    *,
    process_strings: bool = True,
) -> Iterator[ParsedResponse]:
    """Parse a list of DSDataResponse objects into a list of records."""
    for response in responses:
        try:
            parsed_response = parse_response(response, process_strings=process_strings)
        except InvalidResponseError as exc:
            logger.warning("Invalid response, skipping")
            logger.debug(exc)
            continue
        yield parsed_response


def aggregate_responses(responses: Iterable[ParsedResponse]) -> ParsedResponse:
    """Aggregate a list of parsed responses into a single response."""
    parsed_response = ParsedResponse()
    for response in responses:
        parsed_response.records.extend(response.records)
        parsed_response.errors.extend(response.errors)
        parsed_response.meta = parsed_response.meta.merge(response.meta)
    return parsed_response


def parse_response(
    response: DSDataResponse,
    *,
    process_strings: bool = True,
) -> ParsedResponse:
    """Parse a DSDataResponse object into a list of records."""
    dates = response.pydates()
    if dates is None:
        raise InvalidResponseError(
            "Response does not contain dates. Probably the request was invalid."
        )
    meta = parse_meta(response)
    records: RecordDict = collections.defaultdict(dict)
    errors: List[Error] = []
    for data_type_value in response.data_type_values:
        field = data_type_value.data_type
        for symbol_value in data_type_value.symbol_values:
            meta.currencies[symbol_value.symbol][field] = symbol_value.currency
            process_symbol_value(
                records,
                errors,
                field,
                dates,
                symbol_value,
                process_strings=process_strings,
            )
    record_list = []
    for (symbol, date), record in records.items():
        record["symbol"] = symbol
        record["date"] = date
        record_list.append(record)
    return ParsedResponse(record_list, errors, meta)


def process_symbol_value(  # noqa: PLR0913
    records: RecordDict,
    errors: List[Error],
    field: str,
    dates: List[dt.datetime],
    symbol_value: DSSymbolResponseValue,
    *,
    process_strings: bool,
) -> None:
    """Parse a symbol value into a dictionary."""
    is_error = isinstance(symbol_value, DSError)
    if is_error:
        errors.append(Error(field, symbol_value.symbol, symbol_value.value))  # type: ignore[arg-type]
    value = symbol_value.parse()
    if isinstance(symbol_value, DSString) and process_strings:
        value = process_string_value(value)  # type: ignore[arg-type]
    if isinstance(value, list):
        if len(value) != len(dates):
            raise InvalidResponseError(
                "Number of values does not match number of dates."
            )
        for date, xvalue in zip(dates, value):
            records[(symbol_value.symbol, date)][field] = xvalue
    else:
        if len(dates) > 1 and not is_error:
            raise InvalidResponseError("More than one date found for single value.")
        records[(symbol_value.symbol, dates[0])][field] = value


def parse_meta(response: DSDataResponse) -> Meta:
    """Parse meta information from a response."""
    meta = Meta()
    if response.data_type_names:
        meta.data_type_names = {
            kv_pair.key: kv_pair.value for kv_pair in response.data_type_names
        }
    if response.symbol_names:
        meta.symbol_names = {
            kv_pair.key: kv_pair.value for kv_pair in response.symbol_names
        }
    if response.additional_responses:
        meta.additional_responses = {
            kv_pair.key: kv_pair.value for kv_pair in response.additional_responses
        }
    if response.tag:
        meta.tags.append(response.tag)
    return meta


def process_string_value(value: str) -> Optional[Union[str, bool]]:
    """Process a string value further."""
    return {
        "NA": None,
        "N": False,
        "Y": True,
    }.get(value, value)



"""The Datastream Web Service API client package."""

__all__ = (
    "__version__",
    "DSWSClient",
    "DSWSConfig",
    "parse_response",
)

from dsws_client.client import DSWSClient
from dsws_client.config import DSWSConfig
from dsws_client.parse import parse_response
from dsws_client.version import __version__



import datetime as dt
from typing import Any, Union

import msgspec

DateType = Union[dt.date, str]


class DSStringKVPair(msgspec.Struct, rename="pascal", frozen=True):
    """A key-value pair."""

    key: str
    value: Any


class Token(msgspec.Struct, frozen=True):
    """A token."""

    token_value: str
    token_expiry: dt.datetime

    @property
    def is_expired(self) -> bool:
        """Return True if the token is expired."""
        return (self.token_expiry + dt.timedelta(minutes=1)) < dt.datetime.now(
            tz=dt.timezone.utc
        )
